import json
from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket, room_ride, room_car
from greenlet import greenlet

subscribe_ride = {
    'requestId': 'a',
    'type': 'subscribe',
    'room': room_ride,
}
subscribe_car = {
    'requestId': 'b',
    'type': 'subscribe',
    'room': room_car,
}


class TestRoom(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_subcribe_creates_new(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_ride))

        hub = self.client.app.hub

        self.assertEqual(0, len(hub.rooms.keys()))

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws)

        self.assertEqual([room_ride], list(hub.rooms.keys()))

        # Test that a room get a reference to its connections
        # and vice versa

        room = hub.rooms[room_ride]
        connection = room.connections[0]

        self.assertEqual(ws, connection.ws)
        self.assertEqual([room], connection.rooms)

    def test_subscribe_joins_existing(self):
        ws1 = MockWebSocket()
        ws1.mock_incoming_message(json.dumps(subscribe_ride))
        ws2 = MockWebSocket()
        ws2.mock_incoming_message(json.dumps(subscribe_ride))

        hub = self.client.app.hub
        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1)

        self.assertEqual([room_ride], list(hub.rooms.keys()))

        room = hub.rooms[room_ride]
        c1 = room.connections[0]

        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2)
        self.assertEqual([room_ride], list(hub.rooms.keys()))

        c2 = room.connections[1]
        self.assertEqual([c1, c2], room.connections)
        self.assertEqual([room], c1.rooms)
        self.assertEqual([room], c2.rooms)

    def test_close_empty_room(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_ride))

        hub = self.client.app.hub

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws)

        self.assertEqual([room_ride], list(hub.rooms.keys()))

        ws.connection.unsubscribe_all()
        self.assertEqual(0, len(hub.rooms.keys()))

    def test_keeps_open_nonempty_room(self):
        ws1 = MockWebSocket()
        ws1.mock_incoming_message(json.dumps(subscribe_ride))
        ws2 = MockWebSocket()
        ws2.mock_incoming_message(json.dumps(subscribe_ride))
        ws2.mock_incoming_message(json.dumps(subscribe_car))

        hub = self.client.app.hub
        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1)
        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2)

        # Test that both rooms are created and contain
        # the correct connections
        self.assertSetEqual(set([room_ride, room_car]), set(hub.rooms.keys()))
        self.assertEqual([ws2.connection], hub.rooms[room_car].connections)
        self.assertSetEqual(set([ws1.connection, ws2.connection]), set(hub.rooms[room_ride].connections))

        # Close the second websocket, unsubscribing if for all the rooms
        ws2.close()

        # Test that the now empty room_car is closed
        self.assertEqual([room_ride], list(hub.rooms.keys()))

        # And test that the room_ride still contains ws1
        self.assertEqual([ws1.connection], hub.rooms[room_ride].connections)



#!/usr/bin/env python3
from flask_script import Manager
from app import create_app
from settings import Settings
from werkzeug.serving import run_with_reloader
import logging

app = create_app(Settings)
manager = Manager(app)


@manager.command
def runserver():
    from gevent import pywsgi, monkey
    from geventwebsocket.handler import WebSocketHandler
    server = pywsgi.WSGIServer(('', 5000), app, handler_class=WebSocketHandler)

    if app.debug:
        logging.basicConfig(format='%(asctime)s %(message)s')
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        monkey.patch_all()
        run_with_reloader(server.serve_forever)
    else:
        server.serve_forever()


if __name__ == '__main__':
    manager.run()



# For some reason, the test discovery also reads this file
# When the tests loader runs it, it doesn't accept "from app import app":
# SystemError: Parent module '' not loaded, cannot perform relative import

if __name__ == '__main__':
    from gevent import pywsgi, monkey
    from geventwebsocket.handler import WebSocketHandler
    from app import app

    print('Running server')
    server = pywsgi.WSGIServer(('', 5000), app, handler_class=WebSocketHandler)
    monkey.patch_all()
    server.serve_forever()



import json
from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket, room_ride, room_car
from greenlet import greenlet

subscribe_ride = {
    'requestId': 'a',
    'type': 'subscribe',
    'room': room_ride,
}
subscribe_car = {
    'requestId': 'b',
    'type': 'subscribe',
    'room': room_car,
}
publish_ride = {
    'requestId': 'a',
    'type': 'publish',
    'data': [{
        'id': 1,
    }]
}


class TestPublish(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_incoming_trigger_gets_published(self):
        ws1 = MockWebSocket()
        ws1.mock_incoming_message(json.dumps(subscribe_ride))
        ws2 = MockWebSocket()
        ws2.mock_incoming_message(json.dumps(subscribe_ride))
        ws2.mock_incoming_message(json.dumps(subscribe_car))

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1)
        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2)

        # Mock incoming ride trigger
        self.client.flask_test_client.post(
            '/trigger/',
            content_type='application/json',
            data=json.dumps({
                'rooms': [room_ride],
                'data': [{
                    'id': 1,
                }]
            }))

        # The first outgoing message is the allowed_rooms listing
        # The second one is about the successful subscribe
        # The third is the actual publish
        self.assertEqual(3, len(ws1.outgoing_messages))
        self.assertEqual(json.dumps(publish_ride), ws1.outgoing_messages[2])
        # ws2 has 2 subscribe success events
        self.assertEqual(4, len(ws2.outgoing_messages))
        self.assertEqual(json.dumps(publish_ride), ws2.outgoing_messages[3])

        ws2.close()

        self.client.flask_test_client.post(
            '/trigger/',
            content_type='application/json',
            data=json.dumps({
                'rooms': [room_ride],
                'data': [{
                    'id': 1,
                }]
            }))

        self.assertEqual(4, len(ws1.outgoing_messages))
        self.assertEqual(json.dumps(publish_ride), ws1.outgoing_messages[3])

    def test_silent_close(self):
        ws1 = MockWebSocket()
        ws1.mock_incoming_message(json.dumps(subscribe_ride))
        ws2 = MockWebSocket()
        ws2.mock_incoming_message(json.dumps(subscribe_ride))
        ws2.mock_incoming_message(json.dumps(subscribe_car))

        hub = self.client.app.hub

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1)

        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2)

        # Silent close the websocket, the exceptions still exist
        ws2.closed = True

        # Mock incoming ride trigger
        self.client.flask_test_client.post(
            '/trigger/',
            content_type='application/json',
            data=json.dumps({
                'rooms': [room_ride],
                'data': [{
                    'id': 1,
                }]
            }))

        # The first outgoing message is the allowed_rooms listing
        # The second one is about the successful subscribe
        # The third is the actual publish
        self.assertEqual(3, len(ws1.outgoing_messages))
        self.assertEqual(json.dumps(publish_ride), ws1.outgoing_messages[2])

        # Make sure the closed websocket has its connection
        # removed from the rooms
        self.assertEqual([room_ride], list(hub.rooms.keys()))
        self.assertEqual([ws1.connection], hub.rooms[room_ride].connections)



from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket


class TestKeepalive(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_ping_pong(self):
        ws = MockWebSocket()
        ws.mock_incoming_message('ping')
        self.client.open_connection(ws)

        self.assertEqual(2, len(ws.outgoing_messages))
        self.assertEqual('pong', ws.outgoing_messages[1])



from collections import deque
from unittest import mock, TestCase as Case
from geventwebsocket.exceptions import WebSocketError
from geventwebsocket.websocket import MSG_ALREADY_CLOSED
from .hub import Hub
from greenlet import greenlet
import requests


def mock_environ():
    return {
        'HTTP_COOKIE': 'sessionid=foo; csrftoken=bar',
        'HTTP_HOST': 'webapp.test',
        'HTTP_USER_AGENT': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
    }


class TestCase(Case):
    # Reset the Hub
    # Otherwise the hub still has the socket of the previous test
    def tearDown(self):
        if self.client:
            self.client.app.hub = Hub(self.client.app)


class MockWebSocket:
    closed = False
    connection = None

    def __init__(self):
        self.environ = mock_environ()
        self.pending_actions = deque()
        self.outgoing_messages = []

    def send(self, message):
        if self.closed:
            raise WebSocketError(MSG_ALREADY_CLOSED)

        self.outgoing_messages.append(message)

    def close(self):
        self.connection.unsubscribe_all()
        self.closed = True

    def mock_incoming_message(self, msg):
        self.pending_actions.append(self.receive_message(msg))

    def resume_tests(self):
        g_self = greenlet.getcurrent()

        # Switch to the main thread if we are using greenlets.
        # Otherwise just close to jump out of the ws.receive loop
        if g_self.parent:
            g_self.parent.switch()
        else:
            self.close()

    def receive_message(self, msg):
        return msg

    def receive(self):
        result = None
        # Concurrency loop
        # This stack can contain a greenlet switch (method)
        # Or a message receive (str)
        while not result:
            if not len(self.pending_actions):
                self.resume_tests()
                return
            next_action = self.pending_actions.popleft()

            if callable(next_action):
                next_action()
            else:
                result = next_action

        return result


class MockResponse:
    def __init__(self, json_data={}, status_code=200):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


room_ride = '{"target": "ride"}'
room_car = '{"target": "car"}'


def mock_api(url, **kwargs):
    return MockResponse({
            'user': {
                'id': 1,
            },
            'allowed_rooms': [room_ride, room_car],
        }, 200)


class Client:
    def __init__(self, app):
        self.app = app
        self.app.testing = True
        self.flask_test_client = self.app.test_client()
        self._mock_outgoing_requests()

    def __del__(self):
        self._outgoing_requests.stop()

    def open_connection(self, ws, url='/ws/'):
        # We need to invoke a websocket route with the given url
        # No idea why we can't match on just the url_map
        # So we bind it to an empty context.
        context = self.app.wsgi_app.ws.url_map.bind('')

        # Don't really know what the second part in the matched result tuple is
        route = context.match(url)[0]

        route(ws)

    def set_mock_api(self, func):
        self._api_mock.side_effect = func

    def _mock_outgoing_requests(self):
        self._api_mock = mock.MagicMock(side_effect=mock_api)
        self._outgoing_requests = mock.patch.object(requests, 'get', self._api_mock)
        self._outgoing_requests.start()



import uuid
import json


class Connection():
    ws = None
    hub = None
    user_id = None

    def __init__(self, hub, ws):
        self.ws = ws
        self.hub = hub
        self.rooms = []
        self.allowed_rooms = []
        self.uuid = uuid.uuid4()

        # Nasty hack
        try:
            ws.connection = self
        except AttributeError:
            pass

    def parse_user(self, user_data):
        self.user_id = user_data['id']

    def handle_auth_success(self, data):
        self.user_id = data['user']['id']
        self.allowed_rooms = data['allowed_rooms']

        self.send({'allowed_rooms': self.allowed_rooms})

    def handle(self, message):
        if message == 'ping':
            self.ws.send('pong')
            return

        m = json.loads(message)

        requestId = m.get('requestId', None)

        if m['type'] not in ['subscribe']:
            self.send({
                'requestId': requestId,
                'code': 'error',
                'message': 'message-type-not-allowed',
            })

        return self.handle_subscribe(m)

    def handle_subscribe(self, m):
        room = m.get('room', None)
        if room not in self.allowed_rooms:
            self.send({
                'requestId': m['requestId'],
                'code': 'error',
                'message': 'room-not-found',
            })

        room = self.hub.add_to_room(self, m, room)
        self.rooms.append(room)
        self.send({
            'requestId': m['requestId'],
            'code': 'success',
        })

    def unsubscribe_all(self):
        for room in self.rooms:
            room.remove_connection(self)

    def send(self, message):
        if self.ws.closed:
            return

        self.ws.send(json.dumps(message))



import requests
from .testapp.app import app, django_app_url, django_app_port
from high_templar.test import TestCase, Client, MockResponse, MockWebSocket


class TestAuth(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_incoming_websocket_calls_bootstrap(self):
        ws = MockWebSocket()
        self.client.open_connection(ws)
        self.assertEqual(1, requests.get.call_count)

        external_url = 'http://{}:{}/api/bootstrap/'.format(django_app_url, django_app_port)
        self.assertEqual(external_url, requests.get.call_args_list[0][0][0])

    def test_unauth_closed(self):
        def not_authenticated(request, **kwargs):
            return MockResponse({}, 403)

        self.client.set_mock_api(not_authenticated)

        ws = MockWebSocket()
        self.client.open_connection(ws)

        self.assertEqual(True, ws.closed)

    def test_lists_allowed_rooms(self):
        ws = MockWebSocket()
        self.client.open_connection(ws)

        self.assertEqual(['{"allowed_rooms": ["{\\"target\\": \\"ride\\"}", "{\\"target\\": \\"car\\"}"]}'], ws.outgoing_messages)



import json
from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket, room_ride, room_car

subscribe_ride = {
    'requestId': 'a',
    'type': 'subscribe',
    'room': room_ride,
}
subscribe_weird = {
    'requestId': 'b',
    'type': 'subscribe',
    'room': 'weird',
}


class TestSubscribe(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_subscribe_success(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_ride))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'a', 'code': 'success'}, json.loads(ws.outgoing_messages[1]))

    def test_subscribe_unallowed_room(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_weird))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'b', 'code': 'error', 'message': 'room-not-found'}, json.loads(ws.outgoing_messages[1]))



import json
import requests
from flask import make_response
from .connection import Connection


class WebSocketClosedError(Exception):
    pass


class Adapter:
    '''
    Translates incoming requests from the django app to hub actions
    and vice versa.
    '''

    def __init__(self, app):
        self.app = app
        self.base_url = 'http://{}:{}'.format(app.config['DJANGO_APP_URL'], app.config['DJANGO_APP_PORT'])

    def check_auth(self, socket):
        ws = socket.ws
        headers = {
            'cookie': ws.environ['HTTP_COOKIE'],
            'host': ws.environ['HTTP_HOST'],
            'user-agent': ws.environ['HTTP_USER_AGENT']
        }
        res = requests.get('{}/api/bootstrap/'.format(self.base_url), headers=headers)

        if res.status_code != 200:
            return False

        socket.handle_auth_success(res.json())
        return True


class Room:
    '''
    A collection of subscriptions which have access to a scoped permission.

    A permission would be: See actions
    A scoped permission would be: See your own actions

    Then there would "exist" a room for every single user with view_action.own scoped permission
    "exist" between quotation marks, because high-templar only creates a room when someone actually subscribes to it.

    A room keeps track of all connections with the same scoped permission.
    In the case a view_action.own, it would have a maximum of 1 connection.
    '''

    def __init__(self, name, hub):
        self.name = name
        self.hub = hub
        self.subscriptions = []

    def subscribe(self, connection, request):
        sub = Subscription(connection, request)
        self.subscriptions.append(sub)

    # Don't just unsubscribe: remove all subscriptions for a connection
    def remove_connection(self, connection):
        # Clone self.subscriptions because they may be removed when closed
        for s in list(self.subscriptions):
            if s.connection == connection:
                self.subscriptions.remove(s)

        self.close_if_empty()

    def close_if_empty(self):
        if not self.subscriptions:
            self.close()

    def close(self):
        self.hub.close_room(self)

    # Returns a list of closed connections
    def publish(self, data):
        closed_connections = []
        for s in self.subscriptions:
            try:
                s.publish(data)
            except WebSocketClosedError:
                closed_connections.append(s.connection)

        return closed_connections

    @property
    def connections(self):
        return [s.connection for s in self.subscriptions]


class Subscription:
    '''
    A subscription of a connection to a room
    '''

    def __init__(self, connection, request):
        self.connection = connection
        self.requestId = request['requestId']
        self.scope = request.get('scope', {})

    def publish(self, data):
        if self.connection.ws.closed:
            raise WebSocketClosedError

        self.connection.ws.send(json.dumps({
            'requestId': self.requestId,
            'type': 'publish',
            'data': data,
        }))


class Hub:
    '''
    Manages all connections.
    '''

    def __init__(self, app):
        self.app = app
        self.rooms = {}
        self.adapter = Adapter(app)

    def handle_trigger(self, body):
        if 'rooms' not in body:
            return make_response('rooms not specified', 400)
        if 'data' not in body:
            return make_response('data not specified', 400)

        closed_connections = []
        for r in body['rooms']:
            if r not in self.rooms:
                continue
            connections = self.rooms[r].publish(body['data'])
            closed_connections.extend(connections)

        for r_name, room in self.rooms.copy().items():
            for c in closed_connections:
                room.remove_connection(c)

        return make_response('publish success')

    def add_if_auth(self, ws):
        connection = Connection(self, ws)

        auth = self.adapter.check_auth(connection)
        if not auth:
            ws.close()
            return False

        return connection

    def get_or_create_room(self, room_name):
        if room_name in self.rooms:
            return self.rooms[room_name]

        room = Room(room_name, self)
        self.rooms[room_name] = room
        return room

    def add_to_room(self, connection, request, room_name):
        room = self.get_or_create_room(room_name)
        room.subscribe(connection, request)
        return room

    def close_room(self, room):
        # Clone self.rooms because they may be removed when closed
        rooms = self.rooms.copy()
        for r_name, r in rooms.items():
            if r == room:
                del self.rooms[r_name]



from flask_sockets import Sockets
from flask import Flask, request
import logging
import json


def create_app(settings=None):
    app = Flask(__name__)

    if settings:
        app.config.from_object(settings)

    sockets = Sockets(app)

    from .hub import Hub
    app.hub = Hub(app)

    @sockets.route('/ws/')
    def open_socket(ws):
        auth = app.hub.add_if_auth(ws)
        if not auth:
            # todo manage connection close
            return

        connection = auth
        while not connection.ws.closed:
            message = connection.ws.receive()
            if message:
                try:
                    connection.handle(message)
                except Exception as e:
                    logging.error(e, exc_info=True)

    @app.route('/trigger/', methods=['POST'])
    def handle_trigger():
        data = json.loads(request.data.decode())
        return app.hub.handle_trigger(data)

    return app



#!/usr/bin/env python3
from high_templar.main import create_app

django_app_url = 'localhost'
django_app_port = 8000


class Settings:
    DJANGO_APP_URL = django_app_url
    DJANGO_APP_PORT = django_app_port


app = create_app(Settings)



#! /usr/bin/env python3

import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

setup(
    name='high_templar',
    version='0.1.4',
    packages=find_packages(),
    license='MIT',
    description='A python framework for creating a server which handles websockets for an existing API',
    long_description=README,
    author='Jasper Stam',
    author_email='jasper@codeyellow.nl',
    url='https://github.com/CodeYellowBV/high-templar',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    install_requires=[
        'gevent >= 1.1.2',
        'greenlet >= 0.4.12',
        'Flask >= 0.12.0',
        'Flask-Script >= 2.0.5',
        'Flask-Sockets >= 0.2.1',
        'python-dotenv >= 0.6.3',
        'requests >= 2.13.0',
    ],
    test_suite='tests'
)
